"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const Token_1 = __importDefault(require("./type/Token"));
const TokenType_1 = __importDefault(require("./type/TokenType"));
class Lexer {
    constructor(debug = false) {
        this._debug = debug;
    }
    log(state, char, pos, data) {
        if (!this._debug)
            return;
        console.log(`[${state}] "${char}" at ${pos} ${data ? '(' + data + ')' : ''}`);
    }
    static get tokens() {
        return {
            '?': new Token_1.default(TokenType_1.default.Optional, '?', null),
            '+': new Token_1.default(TokenType_1.default.Mutiple, '+', null),
            '{': new Token_1.default(TokenType_1.default.GroupStart, '{', null),
            '}': new Token_1.default(TokenType_1.default.GroupEnd, '}', null),
            '(': new Token_1.default(TokenType_1.default.ArgStart, '(', null),
            ')': new Token_1.default(TokenType_1.default.ArgEnd, ')', null),
            ',': new Token_1.default(TokenType_1.default.Comma, ',', null),
            '!': new Token_1.default(TokenType_1.default.Anonymous, '!', null),
            '|': new Token_1.default(TokenType_1.default.Or, '|', null),
            '&': new Token_1.default(TokenType_1.default.And, '&', null),
            '...': new Token_1.default(TokenType_1.default.Pseudo, '...', null),
            'match': new Token_1.default(TokenType_1.default.Match, 'match', null),
            'group': new Token_1.default(TokenType_1.default.Group, 'group', null),
            'define': new Token_1.default(TokenType_1.default.Define, 'define', null),
            'test': new Token_1.default(TokenType_1.default.Test, 'test', null),
            'if': new Token_1.default(TokenType_1.default.If, 'if', null),
            'else': new Token_1.default(TokenType_1.default.Else, 'else', null),
        };
    }
    // Transform formula to array of statements
    lex(text) {
        let out = [];
        let currPos = 0;
        // Token pos
        let currLine = 1;
        let currColumn = 1;
        function next() {
            currPos++;
            currColumn++;
        }
        function increase(val) {
            currPos += val;
            currColumn += val;
        }
        function nextLine() {
            currLine++;
            currColumn = 1;
        }
        function getPos() {
            return {
                start: {
                    line: currLine,
                    column: currColumn
                }
            };
        }
        // Lookahead text to match string
        function lookaheadString(str) {
            const parts = str.split('');
            for (let i = 0; i < parts.length; i++) {
                if (text[currPos + i] !== parts[i]) {
                    return false;
                }
            }
            return true;
        }
        function lookahead(match, matchNext) {
            const bucket = [];
            while (true) {
                const nextIndex = currPos + bucket.length;
                const nextToken = text[nextIndex];
                if (!nextToken) {
                    break;
                }
                let m = match;
                if (matchNext && bucket.length) {
                    m = matchNext;
                }
                if (m && !m.test(nextToken)) {
                    break;
                }
                bucket.push(nextToken);
            }
            return bucket;
        }
        while (currPos < text.length) {
            const char = text[currPos];
            this.log('LOOP', char, currPos);
            // Skip space
            if (/\s/.test(char)) {
                if (char === '\n') {
                    nextLine();
                }
                next();
                continue;
            }
            // String literal
            if (char === "'" || char === '"') {
                next();
                const bucket = lookahead(new RegExp(`[^${char}]`));
                out.push(new Token_1.default(TokenType_1.default.LiteralString, bucket.join(''), getPos()));
                increase(bucket.length + 1);
                this.log('STRING', char, currPos, bucket.join(''));
                continue;
            }
            // Comment
            if (char === '#') {
                while (currPos < text.length && text[currPos] !== '\n') {
                    next();
                }
            }
            // Name
            const literalRegex = /[a-zA-Z_]/;
            const literalRegexNext = /[a-zA-Z0-9_]/;
            const literalRegexNumber = /[0-9]/;
            if (literalRegex.test(char)) {
                const bucket = lookahead(literalRegex, literalRegexNext);
                let value = bucket.join('');
                increase(bucket.length);
                if (Lexer.tokens[value]) {
                    let tok = Object.assign({}, Lexer.tokens[value]);
                    tok.pos = getPos();
                    out.push(tok);
                }
                else {
                    out.push(new Token_1.default(TokenType_1.default.Name, value, getPos()));
                }
                this.log('NAME', char, currPos, value);
                continue;
            }
            else if (literalRegexNumber.test(char)) {
                const bucket = lookahead(literalRegexNumber);
                let value = bucket.join('');
                increase(bucket.length);
                if (Lexer.tokens[value]) {
                    let tok = Object.assign({}, Lexer.tokens[value]);
                    tok.pos = getPos();
                    out.push(tok);
                }
                else {
                    out.push(new Token_1.default(TokenType_1.default.LiteralNumber, value, { start: { line: currLine, column: currPos } }));
                }
                this.log('NUMBER', char, currPos, value);
                continue;
            }
            // Other tokens
            let didMatch = false;
            for (const key in Lexer.tokens) {
                if (!lookaheadString(key)) {
                    continue;
                }
                if (Lexer.tokens[key]) {
                    let tok = Object.assign({}, Lexer.tokens[key]);
                    tok.pos = getPos();
                    out.push(tok);
                    increase(key.length);
                    didMatch = true;
                }
                else {
                    out.push(new Token_1.default(TokenType_1.default.Unknown, key, getPos()));
                    continue;
                }
            }
            if (didMatch)
                continue;
        }
        return out;
    }
}
exports.default = Lexer;
