"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DefineStatement = exports.IfElseStatement = exports.MatchStatement = exports.GroupStatement = exports.ProgramStatement = void 0;
const StatementType_1 = __importDefault(require("./StatementType"));
class IParentStatement {
    constructor(type, children) {
        this.type = type;
        this.children = children;
    }
}
class ProgramStatement extends IParentStatement {
    constructor(children) {
        super(StatementType_1.default.Program, children);
    }
    serialize(ctx) {
        return this.children.map(child => child.serialize(ctx)).join('');
    }
}
exports.ProgramStatement = ProgramStatement;
function compileParams(original, params) {
    var _a, _b, _c, _d, _e;
    return original
        // Amount
        + (params.amount
            ? ((_a = params.amount) === null || _a === void 0 ? void 0 : _a.max)
                // Min, Max
                ? `{${(_b = params.amount) === null || _b === void 0 ? void 0 : _b.min},${(_c = params.amount) === null || _c === void 0 ? void 0 : _c.max}}`
                // Unlimited
                : ((_d = params.amount) === null || _d === void 0 ? void 0 : _d.min) ? `{${(_e = params.amount) === null || _e === void 0 ? void 0 : _e.min}}` : '+'
            : '')
        // Optional
        + (params.optional ? '?' : '');
}
class GroupStatement extends IParentStatement {
    constructor(children, params, named) {
        super(StatementType_1.default.Group, children);
        this.params = params;
        this.named = named;
    }
    serialize(ctx) {
        let children = this.children.map(child => child.serialize(ctx)).join('');
        let result = this.named === null
            ? `(?:${children})` // Group uncaptured (named=null)
            : this.named === undefined
                ? `(${children})` // Group captured (named=undefined)
                : `(?<${this.named}>${children})`; // Group captured, named
        return compileParams(result, this.params);
    }
}
exports.GroupStatement = GroupStatement;
class IfElseStatement {
    constructor(condition, branchIf, branchElse) {
        this.type = StatementType_1.default.Condition;
        this.condition = condition;
        this.branchIf = branchIf;
        this.branchElse = branchElse;
    }
    serialize(ctx) {
        let condition = this.condition.map(child => child.serialize(ctx)).join('');
        let childrenIf = this.branchIf.map(child => child.serialize(ctx)).join('');
        let childrenElse = this.branchElse.map(child => child.serialize(ctx)).join('');
        return `(?(?=${condition})(${childrenIf}${childrenElse ? ')|(' + childrenElse : ''}))`;
    }
}
exports.IfElseStatement = IfElseStatement;
class DefineStatement {
    constructor(value) {
        this.type = StatementType_1.default.Define;
        this.value = value;
    }
    serialize(ctx) {
        return '';
    }
}
exports.DefineStatement = DefineStatement;
class MatchStatement {
    constructor(values, params = { optional: false }) {
        this.type = StatementType_1.default.Match;
        this.values = values;
        this.params = params;
    }
    serializeOne(ctx, value) {
        // Special
        if (typeof value === 'object' && 'of' in value) {
            // Constants
            switch (value.of) {
                case 'START': {
                    return '^';
                }
                case 'END': {
                    return '$';
                }
                case 'DIGIT': {
                    return '\\d';
                }
                case 'NONDIGIT': {
                    return '\\D';
                }
                case 'LETTER': {
                    return '[a-zA-Z]';
                }
                case 'WORD': {
                    return '\\w';
                }
                case 'NONWORD': {
                    return '\\W';
                }
                case 'ANY': {
                    return '.';
                }
                case 'SPACE': {
                    return '\\s';
                }
            }
            // Definitions
            if (ctx.define[value.of]) {
                return ctx.define[value.of].value.serialize(ctx);
            }
        }
        // Range of characters
        else if (value.startsWith('[') && value.endsWith(']')) {
            return value;
        }
        // Literal
        else {
            return value.replace(/\-/g, '\\-').replace(/\./g, '\\.');
        }
    }
    serialize(ctx) {
        let result = null;
        let doPacking = this.values.length > 1 && this.values.every(value => typeof value === 'string' ? value.length > 1 : true);
        result = this.values.map(value => doPacking ? '(?:' + this.serializeOne(ctx, value) + ')'
            : this.serializeOne(ctx, value)).join('|');
        return result !== null
            ? compileParams(result, this.params)
            : '';
    }
}
exports.MatchStatement = MatchStatement;
